from __future__ import annotations
import socketio
import uvicorn
from tuneflow_py import TuneflowPlugin, Song
from typing import Type
import traceback
from tuneflow_devkit.translate_utils import translate_label
import asyncio
import functools
import json
from tuneflow_devkit.validation_utils import validate_plugin, find_match_plugin_info
from msgpack import unpackb, packb


class Debugger:
    def __init__(self, plugin_class: Type[TuneflowPlugin], bundle_file_path: str) -> None:
        '''
        Creates a local debug server for a single plugin.
        '''

        if plugin_class is None or bundle_file_path is None:
            raise Exception("plugin_class and bundle_file must be provided")
        validate_plugin(plugin_class=plugin_class)
        self._plugin_class = plugin_class
        with open(bundle_file_path, 'r') as bundle_file:
            bundle_info = json.load(bundle_file)
            # Validate plugin and bundle.
            plugin_info = find_match_plugin_info(
                bundle_info=bundle_info, provider_id=plugin_class.provider_id(),
                plugin_id=plugin_class.plugin_id())
            if plugin_info is None:
                raise Exception(
                    "plugin not specified in the bundle, check your bundle.json. For more information checkout https://github.com/tuneflow/tuneflow-py")
            self._plugin_info = plugin_info

        self._daw_sid: None | str = None
        self._sio: None | socketio.AsyncServer = None
        self.port = 18818

    def start(self):
        # create a Socket.IO server
        # Maximum http buffer size set to 100MB.
        sio = socketio.AsyncServer(async_mode='asgi', cors_allowed_origins='*', max_http_buffer_size=1e8)
        self._sio = sio

        async def handle_connect(sid, environ, auth):
            self._daw_sid = sid
            print(
                "===========================================================================")
            print("TuneFlow connected")
            print(
                "===========================================================================")

        async def handle_disconnect(sid):
            self._daw_sid = None
            print(
                "===========================================================================")
            print("TuneFlow disconnected")
            print()
            print(translate_label({
                "en": "IMPORTANT: Please undo the plugin under debugging and re-run it after you restart the DevKit.",
                "zh": "注意: 请在TuneFlow中撤销正在调试的插件，并在DevKit重新启动后重新运行该插件"
            }))
            print(
                "===========================================================================")

        async def handle_get_bundle_info(sid, data):
            return {
                "status": "OK",
                "pluginInfo": self._plugin_info
            }

        def init_plugin_task(plugin_class: Type[TuneflowPlugin], song: Song, sio, sid):
            try:
                params_config = plugin_class.params(song)
                return {"status": "OK",
                        "paramsConfig": params_config,
                        "params": plugin_class._get_default_params(param_config=params_config)
                        }
            except Exception as e:
                print(
                    "=========================== Run Plugin Exception ==========================")
                traceback.print_exc()
                print(
                    "===========================================================================")
                return {
                    "status": "INIT_PLUGIN_EXCEPTION"
                }

        async def handle_init_plugin(sid, data):
            decoded_data = unpackb(data)
            song = Song.deserialize_from_bytestring(decoded_data["song"])
            response = await asyncio.get_event_loop().run_in_executor(None, functools.partial(init_plugin_task, plugin_class=self._plugin_class, song=song, sio=self._sio, sid=self._daw_sid))
            return packb(response)

        def run_plugin_task(plugin_class: Type[TuneflowPlugin], song, params, sio, sid):
            try:
                plugin_class.run(song, params)
            except Exception as e:
                print("================ Run Plugin Exception ================")
                traceback.print_exc()
                print("======================================================")
                return {
                    "status": "RUN_PLUGIN_EXCEPTION"
                }
            return {
                "status": "OK",
                "song": song.serialize_to_bytestring()
            }

        async def handle_run_plugin(sid, data):
            print('run plugin')
            decoded_data = unpackb(data)
            params = decoded_data["params"]
            song = Song.deserialize_from_bytestring(decoded_data["song"])
            result = await asyncio.get_event_loop().run_in_executor(None,  functools.partial(run_plugin_task, plugin_class=self._plugin_class, song=song, params=params, sio=self._sio, sid=self._daw_sid))
            return packb(result)

        sio.on("connect", handle_connect, namespace='/daw')
        sio.on("disconnect", handle_disconnect, namespace='/daw')
        sio.on('get-bundle-info', handle_get_bundle_info, namespace='/daw')
        sio.on('init-plugin', handle_init_plugin, namespace='/daw')
        sio.on('run-plugin', handle_run_plugin, namespace='/daw')

        self._sio = sio

        # Wrap with ASGI application
        self._app = socketio.ASGIApp(self._sio)
        self.print_plugin_info(plugin_info=self._plugin_info)
        print()
        print("======================================================")
        print(translate_label({
            "en": "IMPORTANT: Install this plugin in debug mode from the TuneFlow plugin library panel, then run this plugin from the right-click menu specified in the triggers",
            "zh": "注意: 从 TuneFlow 库中以debug模式安装此插件，随后即可从各级右键菜单中运行此插件"
        }))
        print("======================================================")
        uvicorn.run(self._app, host='127.0.0.1', port=self.port)

    @staticmethod
    def print_plugin_info(plugin_info):
        print(translate_label(
            {"en": "============= Plugin Info =============",
             "zh": "=============== 插件信息 ==============="}))
        print("Provider ID:", plugin_info["providerId"])
        print("Provider Name:", translate_label(
            plugin_info["providerDisplayName"]))
        print("Plugin ID:", plugin_info["pluginId"])
        print("Plugin Name:", translate_label(
            plugin_info["pluginDisplayName"]))
        plugin_description = plugin_info["pluginDescription"]
        print("Plugin Description:", translate_label(plugin_description)
              if plugin_description is not None else 'None')
        print("=======================================")



from tuneflow_py import TuneflowPlugin
from tuneflow_devkit.validation_utils import find_match_plugin_info, validate_plugin
import unittest
import pytest


class TestPluginAnd(unittest.TestCase):
    def test_invalid_provider_id(self):
        class InvalidPlugin1(TuneflowPlugin):
            @staticmethod
            def provider_id() -> str:
                return '0abc'

            @staticmethod
            def plugin_id() -> str:
                return 'abc'

        class InvalidPlugin2(TuneflowPlugin):
            @staticmethod
            def provider_id() -> str:
                return 'a_bc'

            @staticmethod
            def plugin_id() -> str:
                return 'abc'

        bundle_info = {
            "plugins": [
                {
                    "providerId": "0abc",
                    "pluginId": "abc"
                }
            ]
        }

        with pytest.raises(Exception) as e_info_1:
            validate_plugin(plugin_class=InvalidPlugin1)

        self.assertIn("provider_id must only use", e_info_1.value.args[0])

        with pytest.raises(Exception) as e_info_2:
            validate_plugin(plugin_class=InvalidPlugin2)

        self.assertIn("provider_id must only use", e_info_2.value.args[0])

    def test_invalid_plugin_id(self):
        class InvalidPlugin1(TuneflowPlugin):
            @staticmethod
            def provider_id() -> str:
                return 'abc'

            @staticmethod
            def plugin_id() -> str:
                return '0abc'

        class InvalidPlugin2(TuneflowPlugin):
            @staticmethod
            def provider_id() -> str:
                return 'abc'

            @staticmethod
            def plugin_id() -> str:
                return 'a_bc'

        bundle_info = {
            "plugins": [
                {
                    "providerId": "0abc",
                    "pluginId": "abc"
                }
            ]
        }

        with pytest.raises(Exception) as e_info:
            validate_plugin(plugin_class=InvalidPlugin1)

        self.assertIn("plugin_id must only use", e_info.value.args[0])

        with pytest.raises(Exception) as e_info:
            validate_plugin(plugin_class=InvalidPlugin2)

        self.assertIn("plugin_id must only use", e_info.value.args[0])

    def test_missing_in_bundle(self):
        bundle_info = {
            "plugins": [
                {
                    "providerId": "cde",
                    "pluginId": "cde"
                }
            ]
        }
        plugin_info = find_match_plugin_info(bundle_info=bundle_info, provider_id='c_de', plugin_id='cde')
        self.assertIsNone(plugin_info)



from __future__ import annotations
from fastapi import Request, HTTPException, status
from fastapi.testclient import TestClient
from tuneflow_devkit import Runner
from tuneflow_py import Song
from hello_world_plugin import HelloWorldPlugin
import unittest
import pathlib
import json
from typing import Optional
from msgpack import packb, unpackb


class TestPluginCases(unittest.TestCase):
    def test_simple_runner(self):
        bundle_file_path = str(pathlib.PurePath(
            __file__).parent.joinpath('hello_world_plugin.bundle.json'))
        app = Runner(plugin_class_list=[HelloWorldPlugin], bundle_file_path=bundle_file_path).start()

        client = TestClient(app)
        response = client.get("/plugin-bundle-info")
        assert response.status_code == 200
        with open(bundle_file_path, 'r') as bundle_file:
            bundle_info = json.load(bundle_file)
            assert response.json() == bundle_info

    def test_auth_enabled_runner(self):
        bundle_file_path = str(pathlib.PurePath(
            __file__).parent.joinpath('hello_world_plugin.bundle.json'))

        def auth_failed_handler():
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)

        app = Runner(plugin_class_list=[HelloWorldPlugin], bundle_file_path=bundle_file_path).start(config={
            "auth": {
                "handler": auth_failed_handler
            }
        })

        client = TestClient(app)
        response = client.post("/init-plugin-params")
        assert response.status_code == status.HTTP_401_UNAUTHORIZED

    def test_sync_runner(self):
        bundle_file_path = str(pathlib.PurePath(
            __file__).parent.joinpath('hello_world_plugin.bundle.json'))

        app = Runner(plugin_class_list=[HelloWorldPlugin], bundle_file_path=bundle_file_path).start(config={
        })

        client = TestClient(app)
        response = client.post("/jobs", data=packb({
            "providerId": "andantei",
            "pluginId": "hello-world",
            "params": {},
            "song": Song().serialize_to_bytestring()
        }))
        actual_result = response.content
        assert actual_result is not None
        parsed_actual_result = unpackb(actual_result)
        assert parsed_actual_result["status"] == "OK"
        assert parsed_actual_result["song"] is not None

    def test_async_runner(self):
        bundle_file_path = str(pathlib.PurePath(
            __file__).parent.joinpath('hello_world_plugin.bundle.json'))
        global actual_result
        global actual_job_id
        actual_result = None
        actual_job_id = None

        def test_store_uploader(job_id, result):
            global actual_result
            global actual_job_id
            actual_result = result
            actual_job_id = job_id

        def get_result_url(job_id):
            return f"http://download.link/{job_id}"

        app = Runner(plugin_class_list=[HelloWorldPlugin], bundle_file_path=bundle_file_path).start(config={
            "async": {
                "store": {
                    "uploader": test_store_uploader,
                    "resultUrlResolver": get_result_url
                }
            }
        })

        client = TestClient(app)
        accepted_result = client.post("/jobs", data=packb({
            "providerId": "andantei",
            "pluginId": "hello-world",
            "params": {},
            "song": Song().serialize_to_bytestring()
        }))
        parsed_actual_result = unpackb(actual_result)
        assert parsed_actual_result["status"] == "OK"
        assert parsed_actual_result["jobId"] is not None
        assert parsed_actual_result["song"] is not None
        assert accepted_result is not None
        parsed_accepted_result = unpackb(accepted_result.content)
        assert parsed_accepted_result["status"] == "ACCEPTED"
        assert parsed_accepted_result["jobId"] == parsed_actual_result["jobId"]
        assert parsed_accepted_result["resultUrl"] == "http://download.link/{job_id}".format(
            job_id=parsed_actual_result["jobId"])
        assert actual_job_id == parsed_actual_result["jobId"]


if __name__ == '__main__':
    unittest.main()



from typing import Type
from tuneflow_py import TuneflowPlugin
import re


def validate_plugin(plugin_class: Type[TuneflowPlugin]):
    MAX_ID_LENGTH = 100
    if plugin_class.plugin_id() is None:
        raise Exception("plugin_id must be provided")

    if len(plugin_class.plugin_id()) > MAX_ID_LENGTH:
        raise Exception(f"plugin_id must not be longer than {MAX_ID_LENGTH}")

    if re.compile(r'[a-zA-Z]+[0-9a-zA-Z-]*').fullmatch(plugin_class.plugin_id()) is None:
        raise Exception('plugin_id must only use [0-9a-zA-Z-] and the first letter cannot be a digit`')

    if plugin_class.provider_id() is None:
        raise Exception("provider_id must be provided")

    if len(plugin_class.provider_id()) > MAX_ID_LENGTH:
        raise Exception(f"provider_id must not be longer than {MAX_ID_LENGTH}")

    if re.compile(r'[a-zA-Z]+[0-9a-zA-Z-]*').fullmatch(plugin_class.provider_id()) is None:
        raise Exception('provider_id must only use [0-9a-zA-Z-] and the first letter cannot be a digit`')


def find_match_plugin_info(bundle_info, provider_id: str, plugin_id: str):
    for plugin_info in bundle_info["plugins"]:
        if plugin_info["providerId"] == provider_id and plugin_info["pluginId"] == plugin_id:
            return plugin_info
    return None



from __future__ import annotations
from tuneflow_py import TuneflowPlugin, Song
from typing import Type, List
import json
from msgpack import unpackb, packb
from tuneflow_devkit.validation_utils import validate_plugin, find_match_plugin_info
from collections import defaultdict
from fastapi import FastAPI, Request, Response, Depends, BackgroundTasks
from pathlib import PosixPath
import asyncio
import functools
from fastapi.middleware.cors import CORSMiddleware
import traceback
from nanoid import generate as generate_nanoid


class Runner:
    def __init__(self, plugin_class_list: List[Type[TuneflowPlugin]], bundle_file_path: str) -> None:
        '''
        Creates a server for a plugin bundle.

        A plugin bundle can contain multiple plugins in the same virtualenv, their information is specified in the bundle.json.
        '''
        for plugin_class in plugin_class_list:
            validate_plugin(plugin_class=plugin_class)
        self._plugin_class_list = plugin_class_list
        self._plugin_info_map = defaultdict(dict)
        self._bundle_info: dict | None = None
        with open(bundle_file_path, 'r') as bundle_file:
            bundle_info = json.load(bundle_file)
            self._bundle_info = bundle_info
            # Validate plugin and bundle.
            for plugin_class in plugin_class_list:
                provider_id = plugin_class.provider_id()
                plugin_id = plugin_class.plugin_id()
                plugin_info = find_match_plugin_info(
                    bundle_info=bundle_info, provider_id=provider_id,
                    plugin_id=plugin_id)
                if plugin_info is None:
                    raise Exception(
                        "plugin not specified in the bundle, check your bundle.json. For more information checkout https://github.com/tuneflow/tuneflow-py")
                self._plugin_info_map[provider_id][plugin_id] = plugin_info
            # Ensure all entries in bundle.info have corresponding plugin code.
            for plugin_info in bundle_info["plugins"]:
                if not any(
                        plugin_class.provider_id() == plugin_info["providerId"] and plugin_class.plugin_id() ==
                        plugin_info["pluginId"] for plugin_class in plugin_class_list):
                    raise Exception(
                        f'Plugin {plugin_info["providerId"]} {plugin_info["pluginInfo"]} has no corresponding source code in the plugin list')

    def start(self, path_prefix='/', config=None):
        '''
        @param port Port to run the server
        '''

        app = FastAPI()

        cors_allowed_origins = config["corsConfig"]["allowedOrigins"] if config is not None and "corsConfig" in config and "allowedOrigins" in config["corsConfig"] else [
            "*"]
        app.add_middleware(
            CORSMiddleware,
            allow_origins=cors_allowed_origins,
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

        async_config = config["async"] if config and "async" in config else None

        @app.middleware("http")
        async def add_vary_origin_header(request: Request, call_next):
            response = await call_next(request)
            response.headers["Vary"] = 'Origin'
            return response

        get_info_path = str(PosixPath(path_prefix).joinpath('plugin-bundle-info'))
        init_plugin_path = str(PosixPath(path_prefix).joinpath('init-plugin-params'))
        run_plugin_path = str(PosixPath(path_prefix).joinpath('jobs'))

        def init_plugin_task(plugin_class: Type[TuneflowPlugin], song: Song):
            try:
                params_config = plugin_class.params(song)
                return {"status": "OK",
                        "paramsConfig": params_config,
                        "params": plugin_class._get_default_params(param_config=params_config)
                        }
            except Exception as e:
                print(traceback.format_exc())
                return {
                    "status": "ERROR"
                }

        def run_plugin_task(plugin_class: Type[TuneflowPlugin], song, params):
            try:
                plugin_class.run(song, params)
            except Exception as e:
                print(traceback.format_exc())
                return {
                    "status": "ERROR"
                }
            return {
                "status": "OK",
                "song": song.serialize_to_bytestring()
            }

        async def run_plugin_async_task(plugin_class: Type[TuneflowPlugin], song, params, job_id: str, store_uploader):
            # TODO: Revisit to see if we can call run_plugin_task directly.
            response = await asyncio.get_event_loop().run_in_executor(None,  functools.partial(run_plugin_task, plugin_class=plugin_class, song=song, params=params))
            response["jobId"] = job_id
            store_uploader(job_id, packb(response))

        def find_plugin_by_id(plugin_class_list: List[Type[TuneflowPlugin]], provider_id, plugin_id):
            for plugin_class in plugin_class_list:
                if plugin_class.provider_id() == provider_id and plugin_class.plugin_id() == plugin_id:
                    return plugin_class
            raise Exception(f"Cannot find plugin by id {provider_id} {plugin_id}")

        def no_auth_handler():
            pass
        auth_handler = config["auth"]["handler"] if config is not None and "auth" in config and "handler" in config["auth"] else None

        @app.get(get_info_path)
        def handle_get_bundle_info():
            return self._bundle_info

        @app.post(init_plugin_path, dependencies=[Depends(auth_handler if auth_handler else no_auth_handler)])
        async def handle_init_plugin(request: Request):
            raw_body = await request.body()
            body = unpackb(raw_body)
            song = Song.deserialize_from_bytestring(body["song"])
            provider_id = body["providerId"]
            plugin_id = body["pluginId"]
            plugin_class = find_plugin_by_id(self._plugin_class_list, provider_id=provider_id, plugin_id=plugin_id)
            response = await asyncio.get_event_loop().run_in_executor(None, functools.partial(init_plugin_task, plugin_class=plugin_class, song=song))
            return Response(packb(response), headers={"Content-Type": "application/octet-stream"})

        @app.post(run_plugin_path, dependencies=[Depends(auth_handler if auth_handler else no_auth_handler)])
        async def handle_run_plugin(request: Request, background_tasks: BackgroundTasks):
            body = await request.body()
            decoded_data = unpackb(body)
            params = decoded_data["params"]
            provider_id = decoded_data["providerId"]
            plugin_id = decoded_data["pluginId"]
            plugin_class = find_plugin_by_id(self._plugin_class_list, provider_id=provider_id, plugin_id=plugin_id)
            song = Song.deserialize_from_bytestring(decoded_data["song"])
            if async_config:
                # Run in async path.
                job_id = generate_nanoid()
                background_tasks.add_task(
                    run_plugin_async_task, plugin_class=plugin_class, song=song, params=params, job_id=job_id,
                    store_uploader=async_config["store"]["uploader"])
                return Response(packb({
                    "status": "ACCEPTED",
                    "jobId": job_id,
                    "resultUrl": async_config["store"]["resultUrlResolver"](job_id)
                }), headers={"Content-Type": "application/octet-stream"})
            else:
                result = await asyncio.get_event_loop().run_in_executor(None,  functools.partial(run_plugin_task, plugin_class=plugin_class, song=song, params=params))
                return Response(packb(result), headers={"Content-Type": "application/octet-stream"})

        return app



from locale import getlocale
from tuneflow_py import LabelText, Song


def get_system_locale():
    return getlocale()


def get_system_lang_or_default():
    current_locale: str = get_system_locale()[0]
    current_lang = current_locale.split(
        '_')[0] if current_locale is not None else 'en'
    return current_lang


def translate_label(label_text: LabelText):
    if type(label_text) is str:
        return label_text

    current_lang = get_system_lang_or_default()
    label_keys = list(label_text.keys())
    match_locales = [item.split(
        '-')[0].lower() for item in label_keys if item == current_lang]
    match_locale = match_locales[0] if len(match_locales) > 0 else None

    if match_locale is not None:
        return label_text[match_locale]  # type: ignore
    elif len(label_keys) > 0:  # type: ignore
        return label_text[label_keys[0]]  # type: ignore
    else:
        return ''
    